
#include <stdio.h>
#include <stdlib.h>
#include "stm32746g_discovery.h"
#include "stm32746g_discovery_ts.h"
#include "stm32746g_discovery_lcd.h"
#include <stm32746g_discovery_audio.h>
#include "Timer.h"
#include <math.h>
#include "sinetable.h"
#include "freqtable.h"


#define LCD_FRAME_BUFFER          SDRAM_DEVICE_ADDR
#define E_ATTACK 0
#define E_DECAY 1
#define E_SUSTAIN 2
#define E_RELEASE 3
#define E_OFF 4

/* Private function prototypes -----------------------------------------------*/
static void SystemClock_Config(void);
static void CPU_CACHE_Enable(void);
//static void Error_Handler(void);

/* Private functions ---------------------------------------------------------*/
static void UserButton(void);
uint8_t CheckForUserButton(void);
//uint8_t CheckForTouchscreenButton(uint32_t X1, uint32_t X2, uint32_t Y1, uint32_t Y2);




// Sample pragmas to cope with warnings. Please note the related line at
// the end of this function, used to pop the compiler diagnostics status.
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-parameter"
#pragma GCC diagnostic ignored "-Wmissing-declarations"
#pragma GCC diagnostic ignored "-Wreturn-type"





// SAI globals
extern SAI_HandleTypeDef       haudio_out_sai;
extern SAI_HandleTypeDef       haudio_in_sai;

// global variables
TIM_HandleTypeDef TimHandle;
uint32_t uwPrescalerValue = 0;
TS_StateTypeDef  TS_State;
uint8_t FreqStr[256] = {0};
int32_t millisecs, counter;
int32_t time;
uint32_t x, y;
//uint32_t X1, X2, Y1, Y2;
// int buttonstate;


// Sound globals
uint32_t samplerate = 44100;
uint32_t SoundVolume = 80;
int preset;
float phase1, phase2;
int16_t codecBuffer[256];	// 128 samples X 2 channels
void audioBlock(float *input, float *output, int32_t samples);
float inBuffer[256], outBuffer[256]; // interleaved - LRLRLRLRLRLRLRLRLRLRLR - inBuffer[frame << 1] inBuffer[(frame << 1) + 1]
float shape[8192];
float meantone[7] = {0.00272527083333f, 0.00304685416667f, 0.00340658854167f, 0.00364504973958f, 0.00407509747708f, 0.00455638030624f, 0.00509407624166};
int32_t keydown;

// Synthesis types
struct rando
{
	uint32_t seed;
};

struct osco
{
	float pi;	// phase increment
	float phs;	// phase
	float *tab;	// table address
	uint32_t ts;	// table size
	uint32_t tm;	// table mask aka table size minus one
};

typedef struct lino
{
	float val;		// current value
	float dst;		// value to move to
	float inc;		// increment per sample to get from val to dst
} lino;

typedef struct envo
{
	float val;		// current value
	float dst;		// value to move to
	float inc;		// increment per sample to get from val to dst
	float attack;	// in samples (0 to 1)
	float decay;		// in samples (1 to sustain)
	float sustain;	// 0 to 1
	float release;	// in samples (1 to 0)
	int32_t stage; // 0 - attack, 1 - decay, 2 - sustain, 3 -release, 4 - off
} envo;

typedef struct filter0
{
	float c;
	float stt;
}	filter0;

// Synthesis globals
struct osco o[6];
struct filter0 f0;
struct lino amp[3];
envo env[3];

float pitch[3];

/* timer for the sequencer @1kHz */
void timer3_init()
{
	  __HAL_RCC_TIM3_CLK_ENABLE();

	  TimHandle.Instance = TIM3;
	  TimHandle.Init.Period = 10799; // Clock = 216000000/2 = 108000000/(10799+1) = 10000 = 10kHz
	  TimHandle.Init.Prescaler = 9; // = 1kHz clock (starts from zero too!)
	  TimHandle.Init.ClockDivision = 0;
	  TimHandle.Init.CounterMode = TIM_COUNTERMODE_UP;
	  HAL_TIM_Base_Init(&TimHandle) ;
	  HAL_TIM_Base_Start_IT(&TimHandle);

	  /*##-2- Configure the NVIC for TIMx ########################################*/
	  /* Set Interrupt Group Priority */
	  HAL_NVIC_SetPriority(TIM3_IRQn, 4, 0); // lower priority
	  HAL_NVIC_EnableIRQ(TIM3_IRQn);
}


int main(int argc, char* argv[])
{
	int i;
	/* Enable the CPU Cache */
//	CPU_CACHE_Enable();

	HAL_Init();
	SystemClock_Config();
	timer_start();

	// synth init
	for(i = 0; i<8192; i++)
		shape[i] = sinetable[i];
	for (i = 0; i<6; i++)
	{
		o[i].phs = 0.0f;
		o[i].pi = 0.01f;
		o[i].tab = shape;
		o[i].ts = 8192;
		o[i].tm = 8191;
	}
	pitch[0] = meantone[0] * 2.0f;
	pitch[1] = meantone[2] * 2.0f;
	pitch[2] = meantone[4] * 1.0f;

	for(i = 0; i < 3; i++)
	{
		env[i].val = env[i].dst = env[i].inc = 0.0f;
		env[i].stage = E_OFF;
		env[i].attack = 1000.0f;
		env[i].decay = 500.0f;
		env[i].sustain = 0.8f;
		env[i].release = 100000.0f;
	}

	/* LCD Initialization */
	BSP_LCD_Init();

	/* LCD Initialization */
	  BSP_LCD_LayerDefaultInit(0, LCD_FB_START_ADDRESS);

	  /* Enable the LCD */
	  BSP_LCD_DisplayOn();

	  /* Select the LCD Background Layer  */
	  BSP_LCD_SelectLayer(0);

	  /* Clear the Background Layer */
	  BSP_LCD_Clear(LCD_COLOR_WHITE);

	BSP_LCD_DisplayStringAt(0, BSP_LCD_GetFont()->Height, (uint8_t*)"I need an oscillator", CENTER_MODE);
	BSP_LCD_DisplayStringAt(318, BSP_LCD_GetYSize() - 24, (uint8_t *)"Sine      ", LEFT_MODE);
	/* Initialize the touchscreen */
	BSP_TS_Init(BSP_LCD_GetXSize(), BSP_LCD_GetYSize());

	 /* Configure the User Button in GPIO Mode */
	BSP_PB_Init(BUTTON_KEY, BUTTON_MODE_GPIO);

	// output device, volume, sample rate @AUDIO_FREQUENCY_44K
	BSP_AUDIO_OUT_Init(OUTPUT_DEVICE_HEADPHONE, SoundVolume, samplerate);
	// Updates the Audio frame slot configuration.
	BSP_AUDIO_OUT_SetAudioFrameSlot(CODEC_AUDIOFRAME_SLOT_02); // stereo stream out
	// set up first buffer
	BSP_AUDIO_OUT_Play((uint16_t *)codecBuffer, 512);

	// initialize globals
	time = 240;
	millisecs = counter = 0;
	preset = 1;
	// buttonstate = 0;

	// initialize synthesis globals

	timer3_init();

	// Infinite loop
	while (1)
	{
		UserButton();
		Touchscreen();
	}
  // Infinite loop, never return.
}


void audioBlock(float *input, float *output, int32_t frames)
{
	int32_t iphase, i, sample, frame, v;
	float phaseinc1, phaseinc2, phase2mod, mix, voxOut[3];

	o[0].pi = pitch[0];
	o[1].pi = o[0].pi * 1.999f;

	o[2].pi = pitch[1];
	o[3].pi = o[2].pi * 2.01f;

	o[4].pi = pitch[2];
	o[5].pi = o[4].pi * 2.02f;

	for(i = 0; i < frames; i++)
	{
		output[i<<1] = output[(i<<1) + 1] = 0.0f;
		for(v = 0; v < 6; v+=2)
		{
			env[v>>1].val = env[v>>1].val + env[v>>1].inc;
			if(env[v>>1].val > 1.0f) env[v>>1].val = 1.0f;
			if(env[v>>1].val < 0.0f) env[v>>1].val = 0.0f;
			o[v].phs += o[v].pi;
			iphase = (int32_t)(o[v].phs * 8192.0f);	// might overflow if samples is too large.
			voxOut[v>>1]= o[v].tab[iphase & o[v].tm] * env[v>>1].val;

			o[v+1].phs += o[v+1].pi;
			iphase = (int32_t)(o[v+1].phs * 8192.0f);	// might overflow if samples is too large.
			voxOut[v>>1] += o[v+1].tab[iphase & o[v+1].tm] * env[v>>1].val;
			if(keydown == 0)
			{
				switch(env[v>>1].stage)
				{
				case E_RELEASE:
					if(env[v>>1].val <= 0.0f)
					{
						env[v>>1].inc = 0.0f;
						env[v>>1].val = 0.0f;
						env[v>>1].stage = E_OFF;
					}
					break;
				default:
					env[v>>1].stage = E_RELEASE;
					env[v>>1].inc = -1.0f/env[v>>1].release;
					break;
				}
			}
			else
			{
				switch(env[v>>1].stage)
				{
					case E_RELEASE:
					case E_OFF:
						env[v>>1].stage = E_ATTACK;
						env[v>>1].inc = 1.0f/env[v>>1].attack;
						break;
					case E_ATTACK:
						if(env[v>>1].val >= 1.0f)
						{
							env[v>>1].stage = E_DECAY;
							env[v>>1].inc = (env[v>>1].sustain - 1.0f)/env[v>>1].decay;
						}
						break;
					case E_DECAY:
						if(env[v>>1].val <= env[v>>1].sustain)
						{
							env[v>>1].stage = E_SUSTAIN;
							env[v>>1].inc = 0.0f;
							env[v>>1].val = env[v>>1].sustain;
						}
						break;
					case E_SUSTAIN:
						env[v>>1].val = env[v>>1].sustain;
						break;
				}
			}
		}
		output[i<<1] = voxOut[0] * .2f + voxOut[1] * .1f;
		if(output[i<<1] > 1.0f) output[i<<1] = 1.0f;
		if(output[i<<1] < -1.0f) output[i<<1] = -1.0f;
		output[(i<<1) + 1] = voxOut[2] * .2f + voxOut[1] * .1f;
		if(output[(i<<1) + 1] > 1.0f) output[(i<<1) + 1] = 1.0f;
		if(output[(i<<1) + 1] < -1.0f) output[(i<<1) + 1] = -1.0f;
	}

	for(v = 0; v < 6; v++)
	{
		while(o[v].phs >= 1.0f) o[v].phs -= 1.0f;
		while(o[v].phs < 0.0f) o[v].phs += 1.0f;
	}
}




// interrupt from timer 3
void TIM3_IRQHandler(void)
{
	HAL_TIM_IRQHandler(&TimHandle);
}

void HAL_TIM_PeriodElapsedCallback(TIM_HandleTypeDef *htim)
{
}

void AUDIO_OUT_SAIx_DMAx_IRQHandler(void)
{
  HAL_DMA_IRQHandler(haudio_out_sai.hdmatx);
}

void AUDIO_IN_SAIx_DMAx_IRQHandler(void)
{
  HAL_DMA_IRQHandler(haudio_in_sai.hdmarx);
}

void BSP_AUDIO_OUT_HalfTransfer_CallBack(void)
{
	int i;
	audioBlock(inBuffer, outBuffer, 64);
	for(i = 0; i < 128; i+=2)
	{
		codecBuffer[i+0] = (int16_t)((outBuffer[i]) * 32767.0f);
		codecBuffer[i+1] = (int16_t)((outBuffer[i+1]) * 32767.0f);
	}
}

void BSP_AUDIO_OUT_TransferComplete_CallBack(void)
{
	int i;
	audioBlock(inBuffer, outBuffer, 64);
	for(i = 0; i < 128; i+=2)
	{
		codecBuffer[i+128] = (int16_t)((outBuffer[i]) * 32767.0f);
		codecBuffer[i+129] = (int16_t)((outBuffer[i+1]) * 32767.0f);
	}
//	ChangeBuffer is not needed for the F7 BSP code - DMA for Sound (SAI) is in circular (repeat) mode.
//	BSP_AUDIO_OUT_ChangeBuffer((uint16_t *)codecBuffer, 128);
}

void BSP_AUDIO_OUT_Error_CallBack(void)
{
  /* Stop the program with an infinite loop */
  while (1)
  {}
}

void Touchscreen(void)
{
	int x, y, note, octave;
	float phaseincrement;

    BSP_TS_GetState(&TS_State);
    if(TS_State.touchDetected > 0)
    		keydown = 1;
    else
    		keydown = 0;
	x = TS_State.touchX[0];
	y = TS_State.touchY[0];
	note = (x * 22)/BSP_LCD_GetXSize();
	octave = note/7;
	note = note%7;
	phaseincrement = meantone[note]*octave;
	y = (y * 4)/BSP_LCD_GetYSize();
	if(y == 1)
		pitch[0] = phaseincrement;
	if(y == 2)
		pitch[1] = phaseincrement;
	if(y == 3)
		pitch[2] = phaseincrement;
}



static void UserButton(void) // to choose "presets" = partials
{
	int i;
	if (CheckForUserButton())
	{
		preset ++;

		if (preset > 4)
		{
			preset = 1;
		}

		switch ( preset )
		{
		case 1:
			BSP_LCD_DisplayStringAt(318, BSP_LCD_GetYSize() - 24, (uint8_t *)"Sine      ", LEFT_MODE);
			for(i = 0; i<8192; i++)
				shape[i] = sinetable[i];
			break;
		case 2:
			BSP_LCD_DisplayStringAt(318, BSP_LCD_GetYSize() - 24, (uint8_t *)"Saw       ", LEFT_MODE);
			for(i = 0; i<8192; i++)
			{
				shape[i] = sinetable[i];
				shape[i] += sinetable[(i*2)&8191] * 0.5f;
				shape[i] += sinetable[(i*3)&8191] * 0.333333f;
				shape[i] += sinetable[(i*4)&8191] * 0.25f;
				shape[i] += sinetable[(i*5)&8191] * 0.2f;
				shape[i] += sinetable[(i*6)&8191] * 0.166666f;
				shape[i] += sinetable[(i*7)&8191] * 0.142857f;
				shape[i] += sinetable[(i*8)&8191] * 0.125f;
				shape[i] *= 0.55f;
			}
			break;
		case 3:
			BSP_LCD_DisplayStringAt(318, BSP_LCD_GetYSize() - 24, (uint8_t *)"Square    ", LEFT_MODE);
			for(i = 0; i<8192; i++)
			{
				shape[i] = sinetable[i];
				shape[i] += sinetable[(i*3)&8191] * 0.333333f;
				shape[i] += sinetable[(i*5)&8191] * 0.2f;
				shape[i] += sinetable[(i*7)&8191] * 0.142857f;
				shape[i] += sinetable[(i*9)&8191] * 0.111111f;
				shape[i] += sinetable[(i*11)&8191] * 0.090909f;
				shape[i] += sinetable[(i*13)&8191] * 0.076923f;
				shape[i] += sinetable[(i*15)&8191] * 0.066666f;
			}
			break;
		case 4:
			BSP_LCD_DisplayStringAt(318, BSP_LCD_GetYSize() - 24, (uint8_t *)"Triangle   ", LEFT_MODE);
			for(i = 0; i<8192; i++)
			{
				shape[i] = sinetable[i];
				shape[i] -= sinetable[(i*3)&8191] * 0.333333f * 0.333333f;
				shape[i] += sinetable[(i*5)&8191] * 0.2f * 0.2f;
				shape[i] -= sinetable[(i*7)&8191] * 0.142857f * 0.142857f;
				shape[i] += sinetable[(i*9)&8191] * 0.111111f * 0.111111f;
				shape[i] *= 0.75f;
			}
			break;
		}
	}
}

/**
  * @brief  System Clock Configuration
  *         The system Clock is configured as follow :
  *            System Clock source            = PLL (HSE)
  *            SYSCLK(Hz)                     = 216000000
  *            HCLK(Hz)                       = 216000000
  *            AHB Prescaler                  = 1
  *            APB1 Prescaler                 = 4
  *            APB2 Prescaler                 = 2
  *            HSE Frequency(Hz)              = 25000000
  *            PLL_M                          = 25
  *            PLL_N                          = 432
  *            PLL_P                          = 2
  *            PLL_Q                          = 9
  *            VDD(V)                         = 3.3
  *            Main regulator output voltage  = Scale1 mode
  *            Flash Latency(WS)              = 7
  * @param  None
  * @retval None
  */
void SystemClock_Config (void) {
	  RCC_ClkInitTypeDef RCC_ClkInitStruct;
	  RCC_OscInitTypeDef RCC_OscInitStruct;

	  /* Enable HSE Oscillator and activate PLL with HSE as source */
	  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSE;
	  RCC_OscInitStruct.HSEState = RCC_HSE_ON;
	  RCC_OscInitStruct.HSIState = RCC_HSI_OFF;
	  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
	  RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSE;
	  RCC_OscInitStruct.PLL.PLLM = 25;
	  RCC_OscInitStruct.PLL.PLLN = 432;
	  RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV2;
	  RCC_OscInitStruct.PLL.PLLQ = 9;
	  HAL_RCC_OscConfig(&RCC_OscInitStruct);

	  /* Activate the OverDrive to reach the 216 MHz Frequency */
	  HAL_PWREx_EnableOverDrive();


	   // Select PLL as system clock source and configure the HCLK, PCLK1 and PCLK2
	   // clocks dividers
	   RCC_ClkInitStruct.ClockType = (RCC_CLOCKTYPE_SYSCLK | RCC_CLOCKTYPE_HCLK
	       | RCC_CLOCKTYPE_PCLK1 | RCC_CLOCKTYPE_PCLK2);
	   RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
	   // This is expected to work for most large cores.
	   // Check and update it for your own configuration.
	   RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
	   RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV4;
	   RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV2;
	  HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_7);

	  HAL_SYSTICK_Config (HAL_RCC_GetHCLKFreq () / 1000);

	  HAL_SYSTICK_CLKSourceConfig (SYSTICK_CLKSOURCE_HCLK);

	    /* SysTick_IRQn interrupt configuration */
	  HAL_NVIC_SetPriority(SysTick_IRQn, 3, 0);
}

void BSP_LCD_ClockConfig(LTDC_HandleTypeDef *hltdc, void *Params)
{
  static RCC_PeriphCLKInitTypeDef  periph_clk_init_struct;

  /* RK043FN48H LCD clock configuration */
  /* PLLSAI_VCO Input = HSE_VALUE/PLL_M = 1 Mhz */
  /* PLLSAI_VCO Output = PLLSAI_VCO Input * PLLSAIN = 192 Mhz */
  /* PLLLCDCLK = PLLSAI_VCO Output/PLLSAIR = 192/5 = 38.4 Mhz */
  /* LTDC clock frequency = PLLLCDCLK / LTDC_PLLSAI_DIVR_4 = 38.4/4 = 9.6Mhz */
  periph_clk_init_struct.PeriphClockSelection = RCC_PERIPHCLK_LTDC;
  periph_clk_init_struct.PLLSAI.PLLSAIN = 192;
  periph_clk_init_struct.PLLSAI.PLLSAIR = RK043FN48H_FREQUENCY_DIVIDER;
  periph_clk_init_struct.PLLSAIDivR = RCC_PLLSAIDIVR_4;
  HAL_RCCEx_PeriphCLKConfig(&periph_clk_init_struct);
}


uint8_t CheckForUserButton(void)
{
  if (BSP_PB_GetState(BUTTON_KEY) != RESET)
  {
    HAL_Delay(10);
    while (BSP_PB_GetState(BUTTON_KEY) != RESET);
    return 1 ;
  }
  return 0;
}


//uint8_t CheckForTouchscreenButton(uint32_t X1, uint32_t X2, uint32_t Y1, uint32_t Y2)
//{
//  if (((y > Y1 && y < Y2 && x > X1 && x < X2)) != 0)
//  {
//	BSP_TS_GetState(&TS_State);
//	HAL_Delay(10);
//	while (((y > Y1 && y < Y2
//				&& x > X1 && x < X2)) != 0);
//	return 1;
//  }
//  return 0;
//}


/**
  * @brief  CPU L1-Cache enable.
  * @param  None
  * @retval None
  */
static void CPU_CACHE_Enable(void)
{
  /* Enable I-Cache */
  SCB_EnableICache();

  /* Enable D-Cache */
  SCB_EnableDCache();
}

//static void Error_Handler(void)
//{
//  while(1)
//  {
//  }
//}



#pragma GCC diagnostic pop
